import enum


class PTYPE(enum.Enum):
	REQUEST   = 0x00
	PING	  = 0x01
	RESPONSE  = 0x02
	FAULT	 = 0x03
	WORKING   = 0x04
	NOCALL	= 0x05
	REJECT	= 0x06
	ACK	   = 0x07
	CL_CANCEL = 0x08
	FACK	  = 0x09
	CANCELACK = 0x0A
	BIND	  = 0x0B
	BINDACK   = 0x0C
	BINDNAK   = 0x0D
	ALTERCTX  = 0x0E
	ALTERCTX_R= 0x0F
	AUTH3	 = 0x10
	SHUTDOWN  = 0x11
	CO_CANCEL = 0x12
	ORPHANED  = 0x13

class PFC(enum.IntFlag):
	FIRST_FRAG = 0x01
	LAST_FRAG = 0x02
	PENDING_CANCEL = 0x04
	RESERVED_1 = 0x08
	CONC_MPX = 0x10
	DID_NOT_EXECUTE = 0x20
	MAYBE = 0x40
	OBJECT_UUID = 0x80
